import os.path
from typing import Any, Callable, List

import tensorflow as tf
import tensorflow_data_validation as tfdv

get_basename: Callable[[str], str] = os.path.basename


def display_stats(
    dir: str, lhs_split: str = "Split-train", rhs_split: str = "Split-eval"
) -> Any:
    """
    Displays dataset statistics in a notebook cell.

    Two splits of the same dataset can be compared in visualization.
    Consumes `StatisticsGen` output.
    """
    directories: List[str] = tf.io.gfile.glob(os.path.join(dir, "Split-*"))
    names = map(get_basename, directories)
    splits = {
        name: os.path.join(directory, "FeatureStats.pb")
        for name, directory in zip(names, directories)
    }
    return tfdv.visualize_statistics(
        lhs_statistics=tfdv.load_stats_binary(splits[lhs_split]),
        lhs_name=lhs_split,
        rhs_statistics=tfdv.load_stats_binary(splits[rhs_split]),
        rhs_name=rhs_split,
    )


def compare_stats(
    *,
    left_dir: str,
    left_split: str = "train",
    left_label: str = "Train",
    right_dir: str,
    right_split: str = "all",
    right_label: str = "Inference",
) -> Any:
    """
    Displays dataset statistics in a notebook cell.

    Two different datasets can be compared in visualization.
    Consumes `StatisticsGen` output.
    """
    lhs_path = os.path.join(left_dir, f"Split-{left_split}", "FeatureStats.pb")
    lhs_stats = tfdv.load_stats_binary(lhs_path)
    rhs_path = os.path.join(
        right_dir, f"Split-{right_split}", "FeatureStats.pb"
    )
    rhs_stats = tfdv.load_stats_binary(rhs_path)
    return tfdv.visualize_statistics(
        lhs_statistics=lhs_stats,
        lhs_name=left_label,
        rhs_statistics=rhs_stats,
        rhs_name=right_label,
    )



import tfx.v1 as tfx
from absl import logging
from tfx.dsl.component.experimental.decorators import component

from ...visualization.threshold_optimization import g_mean, load_entries


@component  # type:ignore
def BinaryClassificationThresholdOptimizer(
    model_evaluation: tfx.dsl.components.InputArtifact[  # type: ignore
        tfx.types.standard_artifacts.ModelEvaluation
    ],
    model_name: tfx.dsl.components.Parameter[str] = "candidate",  # type: ignore
) -> tfx.dsl.components.OutputDict(best_threshold=float):  # type: ignore
    """
    A component that tries to find the best binary classification threshold
    to use based on `Evaluator` component output.
    """
    entries = load_entries(
        dir=model_evaluation.uri, model_name=model_name, slice_key=None
    )
    entries.sort(key=g_mean, reverse=True)
    best_entry, *_rest = entries
    logging.info(
        "Best threshold found %s with G-mean %f",
        best_entry,
        g_mean(best_entry),
    )
    return {"best_threshold": best_entry.threshold}



import json
import os.path
import tempfile
from types import MappingProxyType
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple

import tfx.v1 as tfx
from absl import logging
from google.cloud import aiplatform
from google.cloud.aiplatform import pipeline_jobs
from pydantic import Field
from tfx.dsl.components.base.base_component import BaseComponent
from tfx.types.channel import Channel

from .base import BasePipelineHelper
from .interface import DEFAULT_CUSTOM_CONFIG, Resources

EMPTY_OVERRIDES: Dict[str, Resources] = {}
DEFAULT_OVERRIDES = MappingProxyType(EMPTY_OVERRIDES)


class VertexAIPipelineHelper(BasePipelineHelper):
    google_cloud_project: str
    """
    The name of GCP project that this pipeline should run in.
    """

    google_cloud_region: str
    """
    The region to run the pipeline in.
    """

    docker_image: str
    """
    The name (location) of the tfx-descended custom docker image built for the project.
    """

    service_account: str
    """
    The name of the service account to use when running the pipeline.
    """

    trainer_machine_type: str = "n1-standard-4"
    """
    The machine type to use for running a training job.
    """

    trainer_accelerator_type: Optional[str] = None
    """
    The (optional) GPU to request for running a training job.
    For example "NVIDIA_TESLA_K80".
    """

    serving_machine_type: str = "n1-standard-4"
    """
    The machine type to use for endpoint serving.
    """

    serving_accelerator_type: Optional[str] = None
    """
    The (optional) GPU to request for serving deployed endpoint.
    For example "NVIDIA_TESLA_K80"
    """

    serving_endpoint_name: str
    """
    The name of the endpoint to serve predictions at.
    """

    num_parallel_trials: int = 1
    """
    Number of hyperparameter tuning trails to execute in parallel.

    The number will indicate how many machines are going to be used at the same time
    for running HP tuning workers.
    """

    resource_overrides: Mapping[str, Resources] = Field(
        default_factory=lambda: DEFAULT_OVERRIDES
    )
    """
    Definition of resources needs of particular components.

    The key is the string ID of a component - which is usually the name of the class
    unless explicitly overridden using `with_id('<new id>')`.

    By default all components are running on `e2-standard-4` machine.
    This setting allows requesting bigger machine sizes if a component needs it.
    The machine type will be automatically selected by Vertex AI to provide at least
    the resources requested here.
    """

    def construct_trainer(
        self,
        *,
        examples: Optional[Channel] = None,
        transform_graph: Optional[Channel] = None,
        schema: Optional[Channel] = None,
        base_model: Optional[Channel] = None,
        hyperparameters: Optional[Channel] = None,
        run_fn: str,
        train_args: Optional[tfx.proto.TrainArgs] = None,
        eval_args: Optional[tfx.proto.EvalArgs] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        machine_spec: Dict[str, Any] = {
            "machine_spec": {
                "machine_type": self.trainer_machine_type,
            },
            "replica_count": 1,
            "container_spec": {
                "image_uri": self.docker_image,
            },
        }
        if self.trainer_accelerator_type is not None:
            machine_spec["machine_spec"][
                "accelerator_type"
            ] = self.trainer_accelerator_type
            machine_spec["machine_spec"]["accelerator_count"] = 1
        vertex_job_spec = {
            "project": self.google_cloud_project,
            "worker_pool_specs": [machine_spec],
        }

        return tfx.extensions.google_cloud_ai_platform.Trainer(
            examples=examples,
            transform_graph=transform_graph,
            schema=schema,
            base_model=base_model,
            hyperparameters=hyperparameters,
            run_fn=run_fn,
            train_args=train_args,
            eval_args=eval_args,
            custom_config={
                **custom_config,
                tfx.extensions.google_cloud_ai_platform.ENABLE_VERTEX_KEY: True,
                tfx.extensions.google_cloud_ai_platform.VERTEX_REGION_KEY: self.google_cloud_region,
                tfx.extensions.google_cloud_ai_platform.TRAINING_ARGS_KEY: vertex_job_spec,
            },
        )

    def construct_tuner(
        self,
        *,
        examples: Channel,
        schema: Optional[Channel] = None,
        transform_graph: Optional[Channel] = None,
        base_model: Optional[Channel] = None,
        tuner_fn: str,
        train_args: Optional[tfx.proto.TrainArgs] = None,
        eval_args: Optional[tfx.proto.EvalArgs] = None,
        tune_args: Optional[tfx.proto.TuneArgs] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        machine_spec: Dict[str, Any] = {
            "machine_spec": {
                "machine_type": self.trainer_machine_type,
            },
            "replica_count": 1,
            "container_spec": {
                "image_uri": self.docker_image,
            },
        }
        if self.trainer_accelerator_type is not None:
            machine_spec["machine_spec"][
                "accelerator_type"
            ] = self.trainer_accelerator_type
            machine_spec["machine_spec"]["accelerator_count"] = 1
        vertex_job_spec = {
            "worker_pool_specs": [
                machine_spec  # replicas will be configured just like master
            ],
        }

        training_inputs = {
            "job_spec": vertex_job_spec,
            "project": self.google_cloud_project,
        }

        return tfx.extensions.google_cloud_ai_platform.Tuner(
            examples=examples,
            transform_graph=transform_graph,
            schema=schema,
            base_model=base_model,
            tuner_fn=tuner_fn,
            train_args=train_args,
            eval_args=eval_args,
            tune_args=tfx.proto.TuneArgs(num_parallel_trials=self.num_parallel_trials),
            custom_config={
                **custom_config,
                tfx.extensions.google_cloud_ai_platform.ENABLE_VERTEX_KEY: True,
                tfx.extensions.google_cloud_ai_platform.VERTEX_REGION_KEY: self.google_cloud_region,
                tfx.extensions.google_cloud_ai_platform.experimental.TUNING_ARGS_KEY: training_inputs,
                tfx.extensions.google_cloud_ai_platform.experimental.REMOTE_TRIALS_WORKING_DIR_KEY: os.path.join(
                    self.pipeline_root, "trials"
                ),
            },
        )

    def construct_pusher(
        self,
        *,
        model: Optional[Channel] = None,
        model_blessing: Optional[Channel] = None,
        infra_blessing: Optional[Channel] = None,
        push_destination: Optional[tfx.proto.PushDestination] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        vertex_serving_spec: Dict[str, Any] = {
            "project_id": self.google_cloud_project,
            "endpoint_name": self.serving_endpoint_name,
            "machine_type": self.serving_machine_type,
        }

        if self.serving_accelerator_type is None:
            serving_image = (
                "europe-docker.pkg.dev/vertex-ai/prediction/tf2-cpu.2-7:latest"
            )
        else:
            serving_image = (
                "europe-docker.pkg.dev/vertex-ai/prediction/tf2-gpu.2-7:latest"
            )
            vertex_serving_spec["accelerator_type"] = self.serving_accelerator_type
            vertex_serving_spec["accelerator_count"] = 1

        return tfx.extensions.google_cloud_ai_platform.Pusher(
            model=model,
            model_blessing=model_blessing,
            infra_blessing=infra_blessing,
            custom_config={
                **custom_config,
                tfx.extensions.google_cloud_ai_platform.ENABLE_VERTEX_KEY: True,
                tfx.extensions.google_cloud_ai_platform.VERTEX_REGION_KEY: self.google_cloud_region,
                tfx.extensions.google_cloud_ai_platform.VERTEX_CONTAINER_IMAGE_URI_KEY: serving_image,
                tfx.extensions.google_cloud_ai_platform.SERVING_ARGS_KEY: vertex_serving_spec,
            },
        )

    def create_and_run_pipeline(
        self,
        components: Iterable[BaseComponent],
        enable_cache: bool = False,
    ) -> None:
        pipeline = tfx.dsl.Pipeline(
            pipeline_name=self.pipeline_name,
            pipeline_root=self.pipeline_root,
            components=list(components),
            enable_cache=enable_cache,
        )

        with tempfile.NamedTemporaryFile(
            "w+", suffix=".json", prefix="pipeline."
        ) as temp_pipeline_file:
            logging.info(
                "Pipeline definition will be stored in temporary file %r",
                temp_pipeline_file.name,
            )
            runner = tfx.orchestration.experimental.KubeflowV2DagRunner(
                config=tfx.orchestration.experimental.KubeflowV2DagRunnerConfig(
                    default_image=self.docker_image
                ),
                output_filename=temp_pipeline_file.name,
            )
            logging.info("Compile VertexAI pipeline definition")
            runner.run(pipeline)  # only writes pipeline definition to file

            # Read the definition and update it with resouce requests
            # (not supported by TFX, but supported by the underlying VertexAI platform)
            logging.info("Patching pipeline definition")
            pipeline_definition = json.load(temp_pipeline_file)
            executors_definition: Dict[str, Dict[str, Any]] = pipeline_definition[
                "pipelineSpec"
            ]["deploymentSpec"]["executors"]
            for component_id, resource in self.resource_overrides.items():
                try:
                    component_definition = executors_definition[
                        f"{component_id}_executor"
                    ]
                except KeyError:
                    logging.warning(
                        "Could not find component %r for resource override %r",
                        component_id,
                        resource,
                    )
                else:
                    logging.info(
                        "Component %r will run with resource overrides %r",
                        component_id,
                        resource,
                    )
                    component_definition["container"]["resources"] = {
                        "cpuLimit": resource.cpu,
                        "memoryLimit": resource.memory,
                    }
            temp_pipeline_file.seek(0)
            temp_pipeline_file.truncate(0)
            json.dump(pipeline_definition, temp_pipeline_file)
            temp_pipeline_file.flush()

            logging.info("Submiting VertexAI pipeline job")
            aiplatform.init(
                project=self.google_cloud_project,
                location=self.google_cloud_region,
            )
            job = pipeline_jobs.PipelineJob(
                template_path=temp_pipeline_file.name, display_name=self.pipeline_name
            )
            job.submit(
                service_account=self.service_account,
            )
            logging.info("Job submitted")



import os.path
from typing import Any

import tensorflow_data_validation as tfdv


def display_schema(dir: str) -> Any:
    """
    Display dataset schema inside a notebook cell.

    Consumes SchemaGen output.
    """
    schema = tfdv.load_schema_text(os.path.join(dir, "schema.pbtxt"))
    return tfdv.display_schema(schema)



import json
import os.path
from typing import Any, Dict

import tensorflow as tf


def display_hyperparams(
    dir: str,
) -> Dict[str, Any]:
    """
    Parse best found hyperparameters.

    Consumes `Tuner` output.
    """
    file_path = os.path.join(dir, "best_hyperparameters.txt")
    with tf.io.gfile.GFile(file_path, "r") as f:
        data = json.load(f)
    hparam_values: Dict[str, Any] = data[
        "values"
    ]  # skip the hparams search space dump
    return hparam_values



from typing import Any, Dict, Optional

import numpy as np
from absl import logging
from matplotlib import pyplot as plt
from matplotlib.colors import Normalize

from .display_metrics import DEFAULT_MODEL_NAME
from .threshold_optimization import load_entries


def plot_binary_classification_confusion_matrix(
    dir: str,
    model_name: str = DEFAULT_MODEL_NAME,
    threshold: float = 0.5,
    slice_key: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Plots binary classification confusion matrix from model evaluation.

    Consumes Evaluator output.
    Uses `matplotlib`.
    """
    # Load evaluation step results
    entries = load_entries(dir=dir, model_name=model_name, slice_key=slice_key)
    # Select the entry with the most matching threshold
    entries.sort(key=lambda entry: abs(entry.threshold - threshold))
    data_at_threshold, *_rest = entries
    threshold_found = data_at_threshold.threshold
    logging.debug(
        "For sought threshold %f found data with threshold %f",
        threshold,
        threshold_found,
    )

    # Extract data points
    tp = data_at_threshold.tp
    fp = data_at_threshold.fp
    fn = data_at_threshold.fn
    tn = data_at_threshold.tn

    # Construct confustion matrix
    sum_actual_positives = tp + fn
    sum_actual_negatives = tn + fp
    count_matrix = np.array([[tp, fn], [fp, tn]])
    logging.debug("Confusion matrix counts %s", count_matrix)
    confusion_matrix = np.array(
        [
            np.array([tp, fn]) / sum_actual_positives,
            np.array([fp, tn]) / sum_actual_negatives,
        ]
    )
    logging.debug("Confusion matrix normalized %s", confusion_matrix)

    # Chart the confusion matrix
    plt.matshow(
        confusion_matrix, cmap="Blues", norm=Normalize(vmin=0.0, vmax=1.0)
    )
    ax = plt.gca()

    value: float
    for (x, y), value in np.ndenumerate(confusion_matrix):
        plt.text(
            y,
            x,
            f"{count_matrix[x, y]} ({value:.1%})",
            va="center",
            ha="center",
            bbox={"boxstyle": "round", "facecolor": "white", "edgecolor": "k"},
        )
    plt.xlabel("Predicted")
    plt.ylabel("Actual")
    plt.title("Confusion Matrix")
    plt.xticks([0, 1], ["positive", "negative"])
    plt.yticks([0, 1], ["positive", "negative"])
    ax.xaxis.set_label_position("top")
    plt.colorbar()



from dataclasses import dataclass
from types import MappingProxyType
from typing import Any, Dict, Iterable, Mapping, Optional

import tfx.v1 as tfx
from tfx.dsl.components.base.base_component import BaseComponent
from tfx.types.channel import Channel
from typing_extensions import Protocol

EMPTY_CONFIG: Dict[str, Any] = {}
DEFAULT_CUSTOM_CONFIG = MappingProxyType(EMPTY_CONFIG)


@dataclass
class Resources:
    """
    Pipeline step resource requirements definition.
    """

    cpu: int
    """Number of CPUs."""

    memory: int
    """Number of RAM memory GB."""


class PipelineHelperInterface(Protocol):
    """
    Describes a tool that hides the complexities of creating TFX components that
    can be run on both local setup and in the cloud.
    """

    def construct_trainer(
        self,
        *,
        examples: Optional[Channel] = None,
        transform_graph: Optional[Channel] = None,
        schema: Optional[Channel] = None,
        base_model: Optional[Channel] = None,
        hyperparameters: Optional[Channel] = None,
        run_fn: str,
        train_args: Optional[tfx.proto.TrainArgs] = None,
        eval_args: Optional[tfx.proto.EvalArgs] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        """Create an appropriate trainer component from given arguments."""

    def construct_tuner(
        self,
        *,
        examples: Channel,
        schema: Optional[Channel] = None,
        transform_graph: Optional[Channel] = None,
        base_model: Optional[Channel] = None,
        tuner_fn: str,
        train_args: Optional[tfx.proto.TrainArgs] = None,
        eval_args: Optional[tfx.proto.EvalArgs] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        """Create an appropriate tuner component from given arguments."""

    def construct_pusher(
        self,
        *,
        model: Optional[Channel] = None,
        model_blessing: Optional[Channel] = None,
        infra_blessing: Optional[Channel] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        """Create an appropriate pusher component from given arguments."""

    def create_and_run_pipeline(
        self,
        components: Iterable[BaseComponent],
        enable_cache: bool = False,
    ) -> None:
        """Create and run a pipeline from given set of components and resources."""

    @property
    def pipeline_root(self) -> str:
        """Get the pipeline artifacts root directory."""
